/*
Module : RelPaths.h
Purpose: Defines the interface for an C++ class to convert to and from relative paths, This is similiar to 
         how Visual Studio stores filenames in a project / workspace. This means that you can copy a 
         project to a different drive and recompile without problems.
Created: PJN / 07-04-2003
History: PJN / 28-12-2006 1. Updated copyright details
                          2. Addition of a CRELATIVEPATHS_EXT_CLASS preprocessor macro to allow the classes to 
                          be more easily added to an extension dll
                          3. Updated the code to clean compile on VC 2005
                          4. Updated the documentation to use the same style as the web site
         PJN / 28-06-2015 1. Updated copyright details.
                          2. Updated the code to clean compile on VC 2008 & later
                          3. Updated the code to require VC 2005 or later
                          4. Updated the code to be a header only implementation
                          5. Reworked the RelativeToAbsolute method to return the string return value as an
                          output parameter instead of a return value.
                          6. Updated the code to include SAL annotations.
                          7. Reworked the code to optionally compile without MFC. By default the code now use 
                          STL classes and idioms but if you define RELATIVEPATHS_MFC_EXTENSIONS the code will 
                          revert back to the MFC behaviour.
                          8. Removed the static critical section member variable of the CRelativePaths class. 
                          Please note that while the RelativeToAbsolute  method is executing it is possible 
                          that code in some other thread of your app could also be modifying the current 
                          working directory. To avoid such race conditions you should serialize that code 
                          with the execution of RelativeToAbsolute with something like a critical section.

Copyright (c) 2003 - 2015 by PJ Naughter (Web: www.naughter.com, Email: pjna@naughter.com)  

All rights reserved.

Copyright / Usage Details:

You are allowed to include the source code in any product (commercial, shareware, freeware or otherwise) 
when your product is released in binary form. You are allowed to modify the source code in any way you want 
except you cannot modify the copyright details at the top of each module. If you want to distribute source 
code with your application, then you are only allowed to distribute versions released by the author. This is 
to maintain a single distribution point for the source code. 

*/


////////////////////// Macros / Defines ///////////////////////////////////////

#pragma once

#ifndef __RELPATHS_H__
#define __RELPATHS_H__

#ifndef CRELATIVEPATHS_EXT_CLASS
#define CRELATIVEPATHS_EXT_CLASS
#endif

#ifndef RELATIVEPATHS_MFC_EXTENSIONS
#ifndef _STRING_
#pragma message("To avoid this message, please put string in your pre compiled header (normally stdafx.h)")
#include <string>
#endif //#ifndef _STRING_
#endif //#ifdef RELATIVEPATHS_MFC_EXTENSIONS

#ifndef __ATLDEF_H__
#pragma message("To avoid this message, please add ATL to your project (normally through stdafx.h)")
#include <atldef.h>
#endif //#ifndef __ATLDEF_H__

#ifndef _In_
#define _In_
#endif //#ifndef _In_

#ifndef _Out_
#define _Out_
#endif //#ifndef _Out_

#ifndef _Success_
#define _Success_(expr)
#endif //#ifndef _Success_


////////////////////// Classes ////////////////////////////////////////////////

class CRELATIVEPATHS_EXT_CLASS CRelativePaths
{
public:
//Typedefs
#ifdef RELATIVEPATHS_MFC_EXTENSIONS
  typedef CString String; 
#else
#ifdef _UNICODE
  typedef std::wstring String;
#else
  typedef std::string String;
#endif //#ifdef _UNICODE
#endif //#ifdef RELATIVEPATHS_MFC_EXTENSIONS

//Methods
  _Success_(return != 0) static BOOL AbsoluteToRelative(_In_ const String& sAbsolutePath, _In_ const String& sRelativeTo, _In_ String& sRelativePath)
  {
    //Set the output parameter to the reasonable default
#ifdef RELATIVEPATHS_MFC_EXTENSIONS
    sRelativePath.Empty();
#else
    sRelativePath.clear();
#endif //#ifdef RELATIVEPATHS_MFC_EXTENSIONS

    //Break up the absolute path into its constituent parts
    TCHAR szAbsoluteDrive[_MAX_DRIVE];
    szAbsoluteDrive[0] = _T('\0');
    TCHAR szAbsoluteDir[_MAX_DIR];
    szAbsoluteDir[0] = _T('\0');
    TCHAR szAbsoluteFname[_MAX_PATH];
    szAbsoluteFname[0] = _T('\0');
    TCHAR szAbsoluteExt[_MAX_EXT];
#ifdef RELATIVEPATHS_MFC_EXTENSIONS
    _tsplitpath_s(sAbsolutePath, szAbsoluteDrive, sizeof(szAbsoluteDrive)/sizeof(TCHAR), szAbsoluteDir, sizeof(szAbsoluteDir)/sizeof(TCHAR),
                  szAbsoluteFname, sizeof(szAbsoluteFname)/sizeof(TCHAR), szAbsoluteExt, sizeof(szAbsoluteExt)/sizeof(TCHAR));
#else
    _tsplitpath_s(sAbsolutePath.c_str(), szAbsoluteDrive, sizeof(szAbsoluteDrive)/sizeof(TCHAR), szAbsoluteDir, sizeof(szAbsoluteDir)/sizeof(TCHAR),
                  szAbsoluteFname, sizeof(szAbsoluteFname)/sizeof(TCHAR), szAbsoluteExt, sizeof(szAbsoluteExt)/sizeof(TCHAR));
#endif //#ifdef RELATIVEPATHS_MFC_EXTENSIONS
    String sAbsoluteDir(szAbsoluteDir);

    //Break up the "sRelativeTo" path into its constituent parts
    TCHAR szRelativeToDrive[_MAX_DRIVE];
    szRelativeToDrive[0] = _T('\0');
    TCHAR szRelativeToDir[_MAX_DIR];
    szRelativeToDir[0] = _T('\0');
  #ifdef RELATIVEPATHS_MFC_EXTENSIONS
    _tsplitpath_s(sRelativeTo, szRelativeToDrive, sizeof(szRelativeToDrive)/sizeof(TCHAR), szRelativeToDir, sizeof(szRelativeToDir)/sizeof(TCHAR), NULL, 0, NULL, 0);
  #else
    _tsplitpath_s(sRelativeTo.c_str(), szRelativeToDrive, sizeof(szRelativeToDrive)/sizeof(TCHAR), szRelativeToDir, sizeof(szRelativeToDir)/sizeof(TCHAR), NULL, 0, NULL, 0);
  #endif //#ifdef RELATIVEPATHS_MFC_EXTENSIONS
    String sRelativeToDir(szRelativeToDir);

    //We can only convert an absolute path to a relative path if the paths are on the same drive or UNC machine 
    if (_tcsicmp(szAbsoluteDrive, szRelativeToDrive) == 0)
    {
      BOOL bMatch = FALSE;
      int nRelativeUpLevel = 0;
      String sRelativeAncestor;
      while (!bMatch)
      {
        sRelativeAncestor = GetFolderAncestor(sRelativeToDir, nRelativeUpLevel);
        if (sRelativeAncestor != _T("\\"))
        {
        #ifdef RELATIVEPATHS_MFC_EXTENSIONS
          ATLASSERT(sRelativeAncestor.GetLength());
          if (sAbsoluteDir.Find(sRelativeAncestor) == 0)
        #else
          ATLASSERT(sRelativeAncestor.length());
          if (sAbsoluteDir.find(sRelativeAncestor) == 0)
        #endif //#ifdef RELATIVEPATHS_MFC_EXTENSIONS
          {
            bMatch = TRUE;
            nRelativeUpLevel = GetAncestorLevel(sRelativeToDir, sRelativeAncestor);

            for (int i=0; i<nRelativeUpLevel; i++)
              sRelativePath += _T("..\\");

          #ifdef RELATIVEPATHS_MFC_EXTENSIONS
            sRelativePath += sAbsoluteDir.Right(sAbsoluteDir.GetLength() - sRelativeAncestor.GetLength() - 1);
          #else
            int nSubLength = static_cast<int>(sAbsoluteDir.length() - sRelativeAncestor.length() - 1);
            sRelativePath += sAbsoluteDir.substr(sAbsoluteDir.length() - nSubLength, nSubLength);
          #endif //#ifdef RELATIVEPATHS_MFC_EXTENSIONS
            sRelativePath += szAbsoluteFname;
            sRelativePath += szAbsoluteExt;
          }
          else
            nRelativeUpLevel++;
        }
        else
        {
          //No matches between the absolute and relative path in that case then the number of levels up
          //for the relative path corresponds to the number of directory levels in sRelativeTo
          bMatch = TRUE;
          nRelativeUpLevel = FoldersUpToRootLevel(sRelativeTo);

          for (int i=0; i<nRelativeUpLevel; i++)
            sRelativePath += _T("..\\");

        #ifdef RELATIVEPATHS_MFC_EXTENSIONS
          sRelativePath += sAbsoluteDir.Right(sAbsoluteDir.GetLength() - sRelativeAncestor.GetLength());
        #else
          int nSubLength = static_cast<int>(sAbsoluteDir.length() - sRelativeAncestor.length());
          sRelativePath += sAbsoluteDir.substr(sAbsoluteDir.length() - nSubLength, nSubLength);
        #endif //#ifdef RELATIVEPATHS_MFC_EXTENSIONS
          sRelativePath += szAbsoluteFname;
          sRelativePath += szAbsoluteExt;
        }
      }
    }
    else
    {
      SetLastError(ERROR_INVALID_PARAMETER);
      return FALSE;
    }

    return TRUE;
  }

  _Success_(return != 0) static BOOL RelativeToAbsolute(_In_ const String& sRelativePath, _In_ const String& sRelativeTo, _Out_ String& sAbsolutePath)
  {
    //Break up the relativeto path into its constituent parts
    TCHAR szRelativeToDrive[_MAX_DRIVE];
    szRelativeToDrive[0] = _T('\0');
    TCHAR szRelativeToDir[_MAX_DIR];
    szRelativeToDir[0] = _T('\0');
  #ifdef RELATIVEPATHS_MFC_EXTENSIONS
    _tsplitpath_s(sRelativeTo, szRelativeToDrive, sizeof(szRelativeToDrive)/sizeof(TCHAR), szRelativeToDir, sizeof(szRelativeToDir)/sizeof(TCHAR), NULL, 0, NULL, 0);
  #else
    _tsplitpath_s(sRelativeTo.c_str(), szRelativeToDrive, sizeof(szRelativeToDrive)/sizeof(TCHAR), szRelativeToDir, sizeof(szRelativeToDir)/sizeof(TCHAR), NULL, 0, NULL, 0);
  #endif //#ifdef RELATIVEPATHS_MFC_EXTENSIONS
    TCHAR szRelativeToDirectory[_MAX_PATH];
    szRelativeToDirectory[0] = _T('\0');
    _tmakepath_s(szRelativeToDirectory, sizeof(szRelativeToDirectory)/sizeof(TCHAR), szRelativeToDrive, szRelativeToDir, NULL, NULL);
    //Remove any trailing "\"
    size_t nRealtiveToDirectoryLen = _tcslen(szRelativeToDirectory) - 1;
    if ((nRealtiveToDirectoryLen > 2) && szRelativeToDirectory[nRealtiveToDirectoryLen] == _T('\\'))
      szRelativeToDirectory[nRealtiveToDirectoryLen] = _T('\0');

    //Only makes sense to convert relative to absolute if sRelativeTo contains drive letter or is a UNC path
    if (_tcslen(szRelativeToDrive) || (_tcsstr(szRelativeToDir, _T("\\\\")) == szRelativeToDir))
    {
      //Remember the old current directory before we change it
      TCHAR szOldCurrentDirectory[_MAX_PATH];
      szOldCurrentDirectory[0] = _T('\0');
      if (GetCurrentDirectory(_MAX_PATH, szOldCurrentDirectory) == 0)
        return FALSE;

      //Temporarily set the current directory to the relativeto directory (Note we deliberatly do not look at the return value from SetCurrentDirectory
      if (SetCurrentDirectory(szRelativeToDirectory))
      {
        TCHAR szAbsolutePath[_MAX_PATH];
        szAbsolutePath[0] = _T('\0'); //by default set it to an empty string

        //Convert the relative path to an absolute path
        LPTSTR szFilePart = NULL;
      #ifdef RELATIVEPATHS_MFC_EXTENSIONS
        if (GetFullPathName(sRelativePath, _MAX_PATH, szAbsolutePath, &szFilePart) == 0)
      #else
        if (GetFullPathName(sRelativePath.c_str(), _MAX_PATH, szAbsolutePath, &szFilePart) == 0)
      #endif //#ifdef RELATIVEPATHS_MFC_EXTENSIONS
        {
          return FALSE;
        }
      #pragma warning(suppress: 6102)
        sAbsolutePath = szAbsolutePath;

        //Restore the current working directory
        SetCurrentDirectory(szOldCurrentDirectory);
      }
      else
      {
        return FALSE;
      }
    }
    else
      sAbsolutePath = sRelativePath;

    return TRUE;
  }

  static String GetFolderAncestor(_In_ const String& sDir, _In_ int nLevel)
  {
    //What will be the return value
    String sAncestor(sDir);

    //Remove the trailing "\" if any
  #ifdef RELATIVEPATHS_MFC_EXTENSIONS
    int nAncestorLength = sAncestor.GetLength();
  #else
    String::size_type nAncestorLength = sAncestor.length();
  #endif //#ifdef RELATIVEPATHS_MFC_EXTENSIONS
    ATLASSERT(nAncestorLength);
    if ((nAncestorLength > 1) && sAncestor[nAncestorLength - 1] == _T('\\'))
    {
    #ifdef RELATIVEPATHS_MFC_EXTENSIONS
      sAncestor = sAncestor.Left(nAncestorLength - 1); 
    #else
      sAncestor = sAncestor.substr(0, nAncestorLength - 1); 
    #endif //#ifdef RELATIVEPATHS_MFC_EXTENSIONS
    }

    //Find the specified amount of folders up
  #ifdef RELATIVEPATHS_MFC_EXTENSIONS
    for (int i=0; i<nLevel && sAncestor.GetLength(); i++)
  #else
    for (int i=0; i<nLevel && sAncestor.length(); i++)
  #endif //#ifdef RELATIVEPATHS_MFC_EXTENSIONS
    {
    #ifdef RELATIVEPATHS_MFC_EXTENSIONS
      int nIndex = sAncestor.ReverseFind(_T('\\'));
      if (nIndex != -1)
    #else
      String::size_type nIndex = sAncestor.find_last_of(_T('\\'));
      if (nIndex != String::npos)
    #endif //#ifdef RELATIVEPATHS_MFC_EXTENSIONS
      {
      #ifdef RELATIVEPATHS_MFC_EXTENSIONS
        sAncestor = sAncestor.Left(nIndex);
        if ((sDir != _T("\\")) && (i == (nLevel - 1)) && (sAncestor.IsEmpty()))
      #else
        sAncestor = sAncestor.substr(0, nIndex);
        if ((sDir != _T("\\")) && (i == (nLevel - 1)) && (sAncestor.length() == 0))
      #endif //#ifdef RELATIVEPATHS_MFC_EXTENSIONS
        {
          sAncestor = _T("\\");
        }
      }
    }

    return sAncestor;
  }

  static int FoldersUpToRootLevel(_In_ const String& sDir)
  {
    String sAncestor(sDir);

    //Remove the trailing "\" if any
  #ifdef RELATIVEPATHS_MFC_EXTENSIONS
    int nAncestorLength = sAncestor.GetLength();
  #else
    String::size_type nAncestorLength = sAncestor.length();
  #endif //#ifdef RELATIVEPATHS_MFC_EXTENSIONS
    ATLASSERT(nAncestorLength);
    if (sAncestor[nAncestorLength - 1] == _T('\\'))
    {
    #ifdef RELATIVEPATHS_MFC_EXTENSIONS
      sAncestor = sAncestor.Left(nAncestorLength - 1);
    #else
      sAncestor = sAncestor.substr(0, nAncestorLength - 1);
    #endif //#ifdef RELATIVEPATHS_MFC_EXTENSIONS
    }

    int nLevels = -1;
  #ifdef RELATIVEPATHS_MFC_EXTENSIONS
    while (sAncestor.GetLength())
  #else
    while (sAncestor.length())
  #endif //#ifdef RELATIVEPATHS_MFC_EXTENSIONS
    {
    #ifdef RELATIVEPATHS_MFC_EXTENSIONS
      int nIndex = sAncestor.ReverseFind(_T('\\'));
      if (nIndex != -1)
    #else
      String::size_type nIndex = sAncestor.find_last_of(_T('\\'));
      if (nIndex != String::npos)
    #endif //#ifdef RELATIVEPATHS_MFC_EXTENSIONS
      {
      #ifdef RELATIVEPATHS_MFC_EXTENSIONS
        sAncestor = sAncestor.Left(nIndex);
      #else
        sAncestor = sAncestor.substr(0, nIndex);
      #endif //#ifdef RELATIVEPATHS_MFC_EXTENSIONS
        nLevels++;
      }
      else
      {
      #ifdef RELATIVEPATHS_MFC_EXTENSIONS
        sAncestor.Empty();
      #else
        sAncestor.clear();
      #endif //#ifdef RELATIVEPATHS_MFC_EXTENSIONS
      }
    }

    return nLevels;
  }
  
  static int GetAncestorLevel(_In_ const String& sDir, _In_ const String& sSubDir)
  {
    //By default assume failure
    int nLevel = -1;

    //What will be the return value
    String sLocalDir(sDir);

  #ifdef RELATIVEPATHS_MFC_EXTENSIONS
    if (sLocalDir.Find(sSubDir) == 0)
  #else
    if (sLocalDir.find(sSubDir) == 0)
  #endif //#ifdef RELATIVEPATHS_MFC_EXTENSIONS
    {
      nLevel = FoldersUpToRootLevel(sDir) - FoldersUpToRootLevel(sSubDir);
    }

    return nLevel;  
  }
};

#endif //#ifndef __RELPATHS_H__
